/***************************************************************************
  qtiffhandler.cpp  : Qt interface to fax tiff files
  ------------------------------------------------------------------
  copyright         : (C) 2002 by Serghei Amelian <serghei@argus.ro>
                    : (C) 2002 by Hans-Peter Jansen <hoj@urpla.net>
  changelog
  ------------------------------------------------------------------
  Thu Aug 29 2002 serghei started
  Mon Dec 16 2002 hp      smooth scaling
  Fri Dec 20 2002 hp      wrapped with sip
  Sat Dec 21 2002 hp      clean up, reorganizations, flicker fix
  Sun Dec 22 2002 hp      default ctor args, status msg handling
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "qtiffhandler.h"
#include <qmessagebox.h>
#include <qwmatrix.h>
#include <qimage.h>
#include <qlabel.h>
#include <qapplication.h>
#include <tiffio.h>
#include <errno.h>
#include <qapplication.h>


QTIFFHandler::QTIFFHandler(QLabel *status, QWidget *parent, const char *name, float zoom, bool smoothScale):
    QObject(parent, name), status(status)
{
    _zoom = zoom;
    _smoothScale = smoothScale;
    _page = _pages = 0;
    _resolution = Normal;
    _degrees = 0.0;
    originalPixmaps.setAutoDelete(true);
    originalImages.setAutoDelete(true);
    _currentPage.resize(0, 0);
    _sender = "";
    _time = "";
}

QTIFFHandler::~QTIFFHandler()
{
}

int QTIFFHandler::load(const char *fileName)
{
    TIFF *tif;

    // open tif file
    if (!(tif = TIFFOpen(fileName, "r"))) {
        error("Failed to open tif file", strerror(errno));
        return 0;
    }

    // clear previous pixmaps
    if (_pages) {
        originalPixmaps.clear();
        originalImages.clear();
        _page = _pages = 0;
    }

    // test type of tiff file
    uint16 bitspersample;
    TIFFGetField(tif, TIFFTAG_BITSPERSAMPLE, &bitspersample);
    if (bitspersample > 1) {
        TIFFClose(tif);
        error("Failed to load tif file", "Can't handle RGB files yet");
        return 0;
    }

    setStatus("Loading page", true);

    // resolution
    float yResolution;
    TIFFGetField(tif, TIFFTAG_YRESOLUTION, &yResolution);
    _resolution = yResolution > 98 ? Fine : Normal;
    TIFFGetField(tif, TIFFTAG_XRESOLUTION, &xResolution);

    // sender
    char *sender;
    TIFFGetField(tif, TIFFTAG_IMAGEDESCRIPTION, &sender);
    char *end = strstr(sender, "close");
    _sender = QString(sender).left(end - sender);

    // time
    char *time;
    TIFFGetField(tif, TIFFTAG_DATETIME, &time);
    end = strstr(time, "close");
    _time = QString(time).left(end - time);

    // read pages
    do {
        setStatus("Loading page " + QString::number(_pages + 1));
        getPage(tif, _resolution);
        _pages++;
    }
    while(TIFFReadDirectory(tif));

    // cleanup
    TIFFClose(tif);

    restoreStatus();
    setPage(1);
    return 1;
}

int QTIFFHandler::getPage(TIFF *tif, Resolution resolution)
{
    // get width and height of page
    uint32 width, height;
    TIFFGetField(tif, TIFFTAG_IMAGEWIDTH, &width);
    TIFFGetField(tif, TIFFTAG_IMAGELENGTH, &height);
    height *= resolution;

    // create buffer
    QImage image;
    image.create(width, height, 1, 2, QImage::BigEndian);
    if (image.isNull())
        return -1;

    image.setColor(0, qRgb(0,0,0));
    image.setColor(1, qRgb(255,255,255));

    // read lines
    uchar** lines = image.jumpTable();
    for (uint row = 0; row < height / resolution; row++) {
        if (TIFFReadScanline(tif, lines[row * _resolution], row, 0) < 0) {
            //TIFFError(__FUNCTION__, "Impossible de lire les lignes");
            //return -1;
        }
        if (resolution == Normal)
            lines[row * 2 + 1] = lines[row * 2];
    }

    // append pixmap to list
    originalPixmaps.append(new QPixmap(image));
    originalImages.append(new QImage(image));

    return 0;
}

void QTIFFHandler::error(const char *strerr, const char *reason)
{
    QMessageBox::critical((QWidget*)this->parent(), "QFaxReader Error",
                  "Error: " + QString(strerr) + ".\nReason: " + QString(reason) + ".");
}

void QTIFFHandler::setStatus(QString msg, bool saveOld)
{
    if (saveOld) {
        QApplication::setOverrideCursor(waitCursor);
        oldStatus = status->text();
    }
    status->setText(msg);
    qApp->processEvents();
    //printf("status: %s, %d\n", msg.latin1(), saveOld);
}

void QTIFFHandler::restoreStatus()
{
    QApplication::restoreOverrideCursor();
    status->setText(oldStatus);
    qApp->processEvents();
    //printf("status restore\n");
}

void QTIFFHandler::close()
{
    _page = _pages = 0;
    _degrees = 0.0;
    _currentPage.resize(0, 0);
    originalPixmaps.clear();
    originalImages.clear();
    emit update();
}

void QTIFFHandler::setPage(int page)
{
    _page = page;
    scale(_zoom);
}

void QTIFFHandler::rotate(double degrees)
{
    _degrees += degrees;
    if (_degrees >= 360.0)
        _degrees -= 360.0;
    scale(_zoom);
}

void QTIFFHandler::scale(float zoomFactor)
{
    if (_pages) {
        setStatus("Scaling page", true);

        if (_smoothScale) {
            QImage image = *originalImages.at(_page - 1);
            int w = int(image.width() * zoomFactor);
            int h = int(image.height() * zoomFactor);
            if (zoomFactor != 1.0) {
                QImage ii = QImage(image).copy();
                ii.invertPixels();
                _currentPage = QPixmap(ii.smoothScale(w, h));
            }
            else { // if (zoomFactor == 1.0)
                _currentPage = QPixmap(image);
            }
        }
        else {
            QWMatrix matrix;
            matrix.scale(zoomFactor, zoomFactor);
            _currentPage = (*originalPixmaps.at(_page - 1)).xForm(matrix);
        }

        //rotate the QPixmap
        if (_degrees) {
            QWMatrix rm;
            rm.rotate(_degrees);
            //qDebug("rotate: %f degrees", _degrees);
            _currentPage = _currentPage.xForm(rm);
        }
        restoreStatus();
    }
    _zoom = zoomFactor;
    emit update();
}

